"""
Migration utilities for Storage Migrations
"""

from abc import ABCMeta, abstractmethod
import logging


log = logging.getLogger(__name__)


class Migration(metaclass=ABCMeta):
    """
    Manager for maintaining various migration actions of the storage: schema, indices, etc
    """

    @property
    @abstractmethod
    def order(self):
        """ Number of this migration in the row of migrations """
        pass

    @abstractmethod
    def up(self):
        """ Migrate DB schema up """
        pass

    @abstractmethod
    def down(self):
        """ Migrate DB schema down """
        pass


class MigrationSet(metaclass=ABCMeta):
    """
    Collection of migrations.
    """

    @abstractmethod
    def migrations(self):
        """
        Get migrations. Subclasses should defile a list of storage migrations here
        """
        return []

    @abstractmethod
    def save_applied_number(self, number):
        """
        Save the last applied up migration number
        """
        pass

    @abstractmethod
    def last_applied(self):
        """
        Number of the last migration that was applied up
        """
        pass

    def _get_migrations(self, number=None, reverse=False):
        """
        Get all sorted migrations or a migration by number
        """
        if number is None:
            return sorted(self.migrations(), key=lambda x: x.order, reverse=reverse)
        return [m for m in self.migrations() if m.order == number]

    def up(self, number=None):
        """
        Runs migrations up. If number was specified, runs particular migration from the set
        """
        for m in self._get_migrations(number, reverse=False):
            if m.order > self.last_applied():
                log.info('Running migration #%i up', m.order)
                m.up()
                self.save_applied_number(m.order)
                log.info('Completed migration #%i up. Last applied is now %i', m.order, m.order)

    def down(self, number=None):
        """
        Runs migrations down. If number was specified, runs particular migration from the set
        """
        for m in self._get_migrations(number, reverse=True):
            if m.order <= self.last_applied():
                log.info('Running migration #%i down', m.order)
                m.down()
                last_applied = m.order-1
                self.save_applied_number(last_applied)
                log.info('Completed migration #%i down. Last applied is now %i', m.order, last_applied)


class Migrator:
    """
    Migrations executor. Just pass a desired set of migrations to it and run up/down.
    If number was specified, runs particular migration
    """
    def __init__(self, migration_set):
        self.migration_set = migration_set

    def up(self, number=None):
        """
        Runs up of a MigrationSet
        """
        self.migration_set.up(number)

    def down(self, number=None):
        """
        Runs down of a MigrationSet
        """
        self.migration_set.down(number)



# pylint: disable=wildcard-import

"""
Collection of Rules for vakt
"""

from .inquiry import *
from .list import *
from .logic import *
from .net import *
from .operator import *
from .string import (
    Equal as StrEqual,
    PairsEqual as StrPairsEqual,
    StartsWith,
    EndsWith,
    Contains,
    RegexMatch,
)



"""
Contains interfaces that all Storages should implement.
"""

from abc import ABCMeta, abstractmethod


class Storage(metaclass=ABCMeta):
    """
    Interface for any storage that persists policies.
    Every storage should implement all the specified methods, but how, it's up to it to decide:
    it can be in-memory storage, SQL database, NoSQL solution, etc.
    """

    @abstractmethod
    def add(self, policy):
        """Store a policy"""
        pass

    @abstractmethod
    def get(self, uid):
        """Retrieve specific policy"""
        pass

    @abstractmethod
    def get_all(self, limit, offset):
        """
        Retrieve all the policies within a window.

        Returns Iterable
        """
        pass

    @abstractmethod
    def find_for_inquiry(self, inquiry, checker=None):
        """
        Get potential policies for a given inquiry.
        Storage is free to decide what policies to return based on the performance and implementation considerations.
        In the worst case - all policies. In the best - policies matched on actions, subjects, resources.
        Mediocre case - match on subject.

        Checker is argument based on the class of provided vakt.checker.*
        This internal checker is responsible for generating the correct type of query to the Storage:
        e.g. RegexChecker, StringExactChecker will result in different queries.

        Returns Iterable
        """
        pass

    @abstractmethod
    def update(self, policy):
        """Update a policy"""
        pass

    @abstractmethod
    def delete(self, uid):
        """Delete a policy"""
        pass

    @staticmethod
    def _check_limit_and_offset(limit, offset):
        if limit < 0:
            raise ValueError("Limit can't be negative")
        if offset < 0:
            raise ValueError("Offset can't be negative")



"""
All Rules relevant to Inquiry context
"""
import warnings

from ..rules.base import Rule

__all__ = [
    'SubjectEqual',
    'ActionEqual',
    'ResourceIn',
]


class SubjectEqual(Rule):
    """
    Rule that is satisfied if the string value equals the Inquiry's Subject.
    For example: context={'user': SubjectEqual()}
    This is an old Rule. Most often operator Rules will be enough.
    """
    def satisfied(self, what, inquiry=None):
        return inquiry and isinstance(what, str) and what == inquiry.subject


class ActionEqual(Rule):
    """
    Rule that is satisfied if the string value equals the Inquiry's Action.
    For example: context={'user': ActionEqual()}
    This is an old Rule. Most often operator Rules will be enough.
    """
    def satisfied(self, what, inquiry=None):
        return inquiry and isinstance(what, str) and what == inquiry.action


class ResourceIn(Rule):
    """
    Rule that is satisfied if list contains the Inquiry's Resource.
    For example: context={'user': ResourceIn()}
    This is an old Rule. Most often list Rules will be enough.
    """
    def satisfied(self, what, inquiry=None):
        return inquiry and isinstance(what, list) and inquiry.resource in what


# Classes marked for removal in next releases
class SubjectEqualRule(SubjectEqual):
    """Deprecated in favor of SubjectEqual"""
    def __init__(self, *args, **kwargs):
        warnings.warn('SubjectEqualRule will be removed in version 2.0. Use SubjectEqual',
                      DeprecationWarning, stacklevel=2)
        super().__init__(*args, **kwargs)


class ActionEqualRule(ActionEqual):
    """Deprecated in favor of ActionEqual"""
    def __init__(self, *args, **kwargs):
        warnings.warn('ActionEqualRule will be removed in version 2.0. Use ActionEqual',
                      DeprecationWarning, stacklevel=2)
        super().__init__(*args, **kwargs)


class ResourceInRule(ResourceIn):
    """Deprecated in favor of ResourceIn"""
    def __init__(self, *args, **kwargs):
        warnings.warn('ResourceInRule will be removed in version 2.0. Use ResourceIn',
                      DeprecationWarning, stacklevel=2)
        super().__init__(*args, **kwargs)
